@@ -1,5 +1,9 @@
 #!/usr/bin/env python3
 
+# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
+#
+# SPDX-License-Identifier: Apache-2.0
+
 import argparse
 import concurrent.futures
 import dataclasses
@@ -9,20 +13,20 @@ import logging
 import os
 import typing
 
-import yaml
-
+import ccc.oci
 import ci.util
 import cnudie.replicate
 import cnudie.retrieve
 import container.util
 import gci.componentmodel as cm
 import product.v2
+import yaml
 
 import ctt.filters as filters
 import ctt.processing_model as processing_model
 import ctt.processors as processors
-from ctt.rbsc_bom import BOMEntry, BOMEntryType, buildAndApplyBOM
 import ctt.uploaders as uploaders
+from ctt.rbsc_bom import BOMEntry, BOMEntryType, buildAndApplyBOM
 
 logger = logging.getLogger(__name__)
 
@@ -264,19 +268,24 @@ def create_jobs(
             )
 
 
-def process_upload_request(upload_request: processing_model.ContainerImageUploadRequest,
-                           upload_mode_images=product.v2.UploadMode.SKIP
-                           ):
+# uploads a single OCI artifact and returns the content digest
+def process_upload_request(
+    upload_request: processing_model.ContainerImageUploadRequest,
+    upload_mode_images=product.v2.UploadMode.SKIP
+) -> str:
     tgt_ref = upload_request.target_ref
 
-    if container.util.image_exists(tgt_ref) and upload_mode_images is product.v2.UploadMode.SKIP:
-        return logger.info(f'{tgt_ref=} exists - skipping processing')
+    oci_client = ccc.oci.oci_client()
+    manifest_blob_ref = oci_client.head_manifest(image_reference=tgt_ref, absent_ok=True)
+    if bool(manifest_blob_ref) and upload_mode_images is product.v2.UploadMode.SKIP:
+        logger.info(f'{tgt_ref=} exists - skipping processing')
+        return manifest_blob_ref.digest
 
     src_ref = upload_request.source_ref
 
     logger.info(f'start processing {src_ref} -> {tgt_ref=}')
 
-    container.util.filter_image(
+    res, _, _ = container.util.filter_image(
         source_ref=src_ref,
         target_ref=tgt_ref,
         remove_files=upload_request.remove_files,
@@ -284,6 +293,41 @@ def process_upload_request(upload_request: processing_model.ContainerImageUpload
 
     logger.info(f'finished processing {src_ref} -> {tgt_ref=}')
 
+    docker_content_digest = res.headers.get('Docker-Content-Digest', None)
+    return docker_content_digest
+
+
+def replace_tag_with_digest(image_reference: str, docker_content_digest: str) -> str:
+    last_part = image_reference.split('/')[-1]
+    if '@' in last_part:
+        src_name, _ = image_reference.rsplit('@', 1)
+    else:
+        src_name, _ = image_reference.rsplit(':', 1)
+
+    return f'{src_name}@{docker_content_digest}'
+
+
+def access_resource_via_digest(res: cm.Resource, docker_content_digest: str) -> cm.Resource:
+    if res.access.type is cm.AccessType.OCI_REGISTRY:
+        digest_ref = replace_tag_with_digest(res.access.imageReference, docker_content_digest)
+        digest_access = cm.OciAccess(
+            cm.AccessType.OCI_REGISTRY,
+            imageReference=digest_ref,
+        )
+    elif res.access.type is cm.AccessType.RELATIVE_OCI_REFERENCE:
+        digest_ref = replace_tag_with_digest(res.access.reference, docker_content_digest)
+        digest_access = cm.RelativeOciAccess(
+            cm.AccessType.RELATIVE_OCI_REFERENCE,
+            reference=digest_ref
+        )
+    else:
+        raise NotImplementedError
+
+    return dataclasses.replace(
+        res,
+        access=digest_access,
+    )
+
 
 def process_images(
     processing_cfg_path,
@@ -293,12 +337,15 @@ def process_images(
     upload_mode=None,
     upload_mode_cd=product.v2.UploadMode.SKIP,
     upload_mode_images=product.v2.UploadMode.SKIP,
+    replace_resource_tags_with_digests=False,
+    skip_cd_validation=False,
 ):
     if processing_mode is ProcessingMode.DRY_RUN:
         ci.util.warning('dry-run: not downloading or uploading any images')
     else:
         logger.info(f'using upload_mode_cd {upload_mode_cd}')
         logger.info(f'using upload_mode_images {upload_mode_images}')
+        logger.info(f'using skip_cd_validation {skip_cd_validation}')
 
     if upload_mode_images is product.v2.UploadMode.FAIL:
         raise NotImplementedError('upload-mode-image=fail is not a valid argument.')
@@ -325,7 +372,22 @@ def process_images(
     def process_job(processing_job: processing_model.ProcessingJob):
         # do actual processing
         if processing_mode is ProcessingMode.REGULAR:
-            process_upload_request(processing_job.upload_request, upload_mode_images)
+            docker_content_digest = process_upload_request(processing_job.upload_request, upload_mode_images)
+
+            if replace_resource_tags_with_digests:
+                if not docker_content_digest:
+                    raise RuntimeError(f'No Docker_Content_Digest returned for {processing_job=}')
+
+                processing_job.upload_request = dataclasses.replace(
+                    processing_job.upload_request,
+                    target_ref=replace_tag_with_digest(processing_job.upload_request.target_ref, docker_content_digest),
+                )
+
+                if processing_job.processed_resource:
+                    processing_job.processed_resource = access_resource_via_digest(processing_job.processed_resource, docker_content_digest)
+                else:
+                    processing_job.resource = access_resource_via_digest(processing_job.resource, docker_content_digest)
+
             bom_resources.append(
                 BOMEntry(
                     processing_job.upload_request.target_ref,
@@ -370,7 +432,8 @@ def process_images(
         # patch-in overwrites (caveat: must be done sequentially, as lists are not threadsafe)
         for job in job_group:
             component = job.component
-            patched_resources[job.resource.identity(component.resources)] = job.processed_resource or job.resource
+            patched_resource = job.processed_resource or job.resource
+            patched_resources[job.resource.identity(component.resources)] = patched_resource
             continue
 
         res_list = []
@@ -413,6 +476,16 @@ def process_images(
             component=component,
         )
 
+        # Validate the patched component-descriptor and exit on fail
+        if not skip_cd_validation:
+            try:
+                dict_cd = yaml.safe_load(yaml.dump(data=dataclasses.asdict(component_descriptor), Dumper=cm.EnumValueYamlDumper))
+                cm.ComponentDescriptor.validate(dict_cd, validation_mode=cm.ValidationMode.FAIL)
+            except Exception as e:
+                logger.error(f'Schema validation for component-descriptor '
+                             f'{component_descriptor.component.name}:{component_descriptor.component.version} failed with {e}')
+                raise e
+
         src_ctx_repo_base_url = component_descriptor.component.repositoryContexts[-2].baseUrl
         if processing_mode is ProcessingMode.REGULAR:
             if component.name == source_comp.name and component.version == source_comp.version:
@@ -484,6 +557,7 @@ def main():
     parser.add_argument('-s', '--src-ctx-repo-url')
     parser.add_argument('-n', '--component-name')
     parser.add_argument('-v', '--component-version')
+    parser.add_argument('-l', '--skip-cd-validation', action='store_true')
     parser.add_argument('-g', '--rbsc-git-url')
     parser.add_argument('-b', '--rbsc-git-branch')
     parser.add_argument('-u', '--upload-mode-cd',
@@ -498,6 +572,9 @@ def main():
                         ],
                         default=product.v2.UploadMode.SKIP.value
                         )
+    parser.add_argument('-r', '--replace-resource-tags-with-digests', action='store_true',
+                        help='replace tags with digests for resources that are accessed via OCI references'
+                        )
 
     parsed = parser.parse_args()
 
@@ -539,6 +616,8 @@ def main():
         processing_mode=processing_mode,
         upload_mode_cd=product.v2.UploadMode(parsed.upload_mode_cd),
         upload_mode_images=product.v2.UploadMode(parsed.upload_mode_images),
+        replace_resource_tags_with_digests=parsed.replace_resource_tags_with_digests,
+        skip_cd_validation=parsed.skip_cd_validation,
     )
 
     if parsed.component_descriptor:
